from __future__ import unicode_literals
import sys
import ctypes
import ctypes.util

if sys.version < '3.4':
    import imp

    def module_exists(name):
        try:
            imp.find_module(name)
            return True
        except ImportError:
            return False
else:
    import importlib.util

    def module_exists(name):
        if importlib.util.find_spec(name):
            return True
        else:
            return False


def config():
    if sys.platform.startswith("win"):
        return

    libdl = ctypes.CDLL(ctypes.util.find_library('dl'))
    libc = ctypes.PyDLL(None)

    class Dl_info(ctypes.Structure):
        _fields_ = (('dli_fname', ctypes.c_char_p),
                    ('dli_fbase', ctypes.c_void_p),
                    ('dli_sname', ctypes.c_char_p),
                    ('dli_saddr', ctypes.c_void_p))

    libdl.dladdr.argtypes = (ctypes.c_void_p, ctypes.POINTER(Dl_info))
    info = Dl_info()
    libdl.dladdr(libc.Py_IsInitialized, ctypes.byref(info))

    python = sys.executable
    libpython = info.dli_fname.decode()

    try:
        lib = ctypes.PyDLL(libpython)
    except OSError:
        libpython = ""
    else:
        if lib.Py_IsInitialized() == 0:
            libpython = ""

    try:
        import sysconfig
        pythonhome = sysconfig.get_config_vars('prefix')[0] + ":" + \
            sysconfig.get_config_vars('exec_prefix')[0]
    except Exception:
        pythonhome = None
    numpy = module_exists("numpy")

    return (python, libpython, pythonhome, numpy)



from __future__ import unicode_literals

from six import StringIO
from contextlib import contextmanager


output_buffer = StringIO()
error_buffer = StringIO()
_flushable = True
_capture_state = False
_callback = None


def reg_callback(callback):
    global _callback
    _callback = callback


def write_console(buf, otype):
    if _capture_state:
        if otype == 0:
            output_buffer.write(buf)
        else:
            error_buffer.write(buf)
    else:
        _callback.write_console_ex(buf, otype)


def read_buffer(b):
    b.seek(0)
    out = b.getvalue()
    b.seek(0)
    b.truncate(0)
    return out


def read_stdout():
    return read_buffer(output_buffer)


def flush_stdout():
    if not _flushable:
        return
    out = read_stdout()
    if out:
        _callback.write_console_ex(out, 0)


def read_stderr():
    return read_buffer(error_buffer)


def flush_stderr():
    if not _flushable:
        return
    err = read_stderr()
    if err:
        _callback.write_console_ex(err, 1)


def flush():
    flush_stdout()
    flush_stderr()


@contextmanager
def capture_console(flushable=True):
    global _capture_state
    global _flushable
    _capture_state_old = _capture_state
    _capture_state = True
    _flushable_old = _flushable
    _flushable = flushable and _flushable
    try:
        yield
    finally:
        _capture_state = _capture_state_old
        if flushable and _capture_state == 0:
            flush()
        _flushable = _flushable_old



from __future__ import unicode_literals
import os
import re
import subprocess
import sys
import ctypes
from distutils.version import LooseVersion
if sys.version_info[0] >= 3:
    from shutil import which
else:
    from backports.shutil_which import which

if sys.platform.startswith('win'):
    if sys.version_info[0] >= 3:
        from winreg import OpenKey, QueryValueEx, HKEY_LOCAL_MACHINE, KEY_READ
    else:
        from _winreg import OpenKey, QueryValueEx, HKEY_LOCAL_MACHINE, KEY_READ


def read_registry(key, valueex):
    reg_key = OpenKey(HKEY_LOCAL_MACHINE, key, 0, KEY_READ)
    return QueryValueEx(reg_key, valueex)


def getRhome(path, throw=False):
    rhome = ""

    if sys.platform.startswith("win") and path and not path.endswith(".exe"):
        path = path + ".exe"

    if not which(path):
        return None
    try:
        rhome = subprocess.check_output([path, "RHOME"]).decode("utf-8").strip()
    except Exception:
        rhome = None

    return rhome


def verify_Rhome(rhome):
    if sys.platform.startswith("win"):
        path = os.path.join(rhome, "bin", "x64" if sys.maxsize > 2**32 else "i386", "R.dll")
    elif sys.platform == "darwin":
        path = os.path.join(rhome, "lib", "libR.dylib")
    else:
        path = os.path.join(rhome, "lib", "libR.so")

    if not os.path.exists(path):
        if sys.platform.startswith("win"):
            another_path = os.path.join(
                rhome, "bin", "i386" if sys.maxsize > 2**32 else "x64", "R.dll")
            if os.path.exists(another_path):
                raise RuntimeError("R and python architectures do not match.")
        raise RuntimeError("R share library ({}) does not exist.".format(path))


def Rhome():
    rhome = None

    if 'R_BINARY' in os.environ:
        rhome = getRhome(os.environ['R_BINARY'], throw=True)
        if not rhome:
            raise RuntimeError("R binary ({}) does not exist.".format(os.environ['R_BINARY']))

    if not rhome and 'R_HOME' in os.environ:
        rhome = os.environ['R_HOME']
        if not os.path.isdir(rhome):
            raise RuntimeError("R_HOME ({}) does not exist.".format(rhome))
        return rhome

    if not rhome:
        rhome = getRhome("R")

    try:
        if sys.platform.startswith("win") and not rhome:
            rhome = read_registry("Software\\R-Core\\R", "InstallPath")[0]
    except Exception:
        rhome = ""

    if rhome:
        os.environ['R_HOME'] = rhome
    else:
        raise RuntimeError("Cannot determine R HOME.")

    verify_Rhome(rhome)

    return rhome


def ensure_path(rhome=None):
    if not rhome:
        rhome = Rhome()
    if sys.platform.startswith("win"):
        libRdir = os.path.join(rhome, "bin", "x64" if sys.maxsize > 2**32 else "i386")

        # make sure Rblas.dll can be reached
        try:
            msvcrt = ctypes.cdll.msvcrt
            msvcrt._wgetenv.restype = ctypes.c_wchar_p
            path = msvcrt._wgetenv(ctypes.c_wchar_p("PATH"))
            if libRdir not in path:
                path = libRdir + ";" + path
                msvcrt._wputenv(ctypes.c_wchar_p("PATH={}".format(path)))
        except Exception as e:
            print(e)
            pass


def rversion(rhome=None):
    if not rhome:
        rhome = Rhome()
    try:
        output = subprocess.check_output(
            [os.path.join(rhome, "bin", "R"), "--slave", "-e", "cat(as.character(getRversion()))"],
            stderr=subprocess.STDOUT).decode("utf-8").strip()
        version = LooseVersion(output)
    except Exception:
        version = LooseVersion("1000.0.0")
    return version


UTFPATTERN = re.compile(b"\x02\xff\xfe(.*?)\x03\xff\xfe", re.S)
if sys.version_info[0] >= 3:
    DECODE_ERROR_HANDLER = "backslashreplace"
else:
    DECODE_ERROR_HANDLER = "replace"


def rconsole2str(buf):
    ret = ""
    m = UTFPATTERN.search(buf)
    while m:
        a, b = m.span()
        ret += system2utf8(buf[:a]) + m.group(1).decode("utf-8", DECODE_ERROR_HANDLER)
        buf = buf[b:]
        m = UTFPATTERN.search(buf)
    ret += system2utf8(buf)
    return ret


if sys.platform == "win32":
    """
    The following only works after setlocale in C and
    R will initialize it for us. To mimic the behaviour, consider
    ```
    ctypes.cdll.msvcrt.setlocale(0, ctypes.c_char_p("chinese-traditional"))
    ```
    """

    mbtowc = ctypes.cdll.msvcrt.mbtowc
    mbtowc.argtypes = [
        ctypes.POINTER(ctypes.c_wchar),
        ctypes.POINTER(ctypes.c_char),
        ctypes.c_size_t]
    mbtowc.restype = ctypes.c_int

    wctomb = ctypes.cdll.msvcrt.wctomb
    wctomb.argtypes = [ctypes.POINTER(ctypes.c_char), ctypes.c_wchar]
    wctomb.restype = ctypes.c_int

    def system2utf8(buf):
        wcbuf = ctypes.create_unicode_buffer(1)
        text = ""
        while buf:
            n = mbtowc(wcbuf, buf, len(buf))
            if n <= 0:
                break
            text += wcbuf[0]
            buf = buf[n:]
        return text

    def utf8tosystem(text):
        s = ctypes.create_string_buffer(10)
        buf = b""
        for c in text:
            try:
                n = wctomb(s, c)
            except Exception:
                n = -1

            if n > 0:
                buf += s[:n]
            else:
                buf += "\\u{{{}}}".format(hex(ord(c))[2:]).encode("ascii")
        return buf

else:
    def system2utf8(buf):
        return buf.decode("utf-8", DECODE_ERROR_HANDLER)

    def utf8tosystem(text):
        return text.encode("utf-8", "backslashreplace")


def id_str(x):
    return str(id(x))



import os
import re
import sys
from cffi import FFI
ffibuilder = FFI()

cwd = os.path.dirname(os.path.realpath(__file__))
cdef_pattern = re.compile("// begin cdef([^$]*)// end cdef")
cb_cdef_pattern = re.compile("// begin cb cdef([^$]*)// end cb cdef")


for header_file in ["R.h", "libR.h", "gli.h", "parse.h", "process_events.h"]:
    with open(os.path.join(cwd, header_file), "r") as f:
        m = cdef_pattern.search(f.read(), re.M)
        ffibuilder.cdef(m.group(1).replace("RAPI_EXTERN", "extern"))

if sys.platform.startswith("win"):
    ffibuilder.cdef("""
        extern char *(*get_R_HOME)(void);
        extern char *(*getRUser)(void);
        extern int* UserBreak_t;
        extern int* CharacterMode_t;
        extern int* EmitEmbeddedUTF8_t;
        extern int (*GA_peekevent)(void);
        extern int (*GA_initapp)(int, char **);
    """)
else:
    ffibuilder.cdef("""
        extern void* (*R_InputHandlers);
        extern void (**R_PolledEvents_t)(void);
        extern void* (*R_checkActivity)(int usec, int ignore_stdin);
        extern void (*R_runHandlers)(void* handlers, void* mask);
        extern int* R_interrupts_pending_t;
    """)

with open(os.path.join(cwd, "libR.h"), "r") as f:
    m = cb_cdef_pattern.search(f.read(), re.M)
    ffibuilder.cdef("""
        extern "Python+C" {{
            {}
        }}
    """.format(m.group(1)))

ffibuilder.set_source(
    "rchitect._cffi",
    """
    # include "gli.h"
    # include "libR.h"
    # include "parse.h"
    # include "process_events.h"
    """,
    include_dirs=['rchitect/src'],
    sources=[
        'rchitect/src/libR.c',
        'rchitect/src/gli.c',
        'rchitect/src/parse.c',
        'rchitect/src/process_events.c'
    ])

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



from __future__ import unicode_literals
import os
from rchitect.interface import rcall, rcopy, set_hook, package_event


def configure():
    def _configure():
        rcall(
            ("base", "source"),
            os.path.join(os.path.dirname(__file__), "config.R"),
            rcall(("base", "new.env")))

    if "reticulate" in rcopy(rcall(("base", "loadedNamespaces"))):
        _configure()
    else:
        set_hook(package_event("reticulate", "onLoad"), lambda x, y: _configure())



from __future__ import unicode_literals

from rchitect._cffi import ffi, lib
from .types import box, unbox


_global_set = dict()


@ffi.def_extern()
def xptr_finalizer(s):
    h = lib.R_ExternalPtrAddr(s)
    del _global_set[h]


def new_xptr_p(x):
    h = ffi.new_handle(x)
    hp = ffi.cast("void*", h)
    _global_set[hp] = h
    s = lib.R_MakeExternalPtr(hp, lib.R_NilValue, lib.R_NilValue)
    lib.R_RegisterCFinalizerEx(s, ffi.addressof(lib, "xptr_finalizer"), 1)
    return s


def new_xptr(x):
    return box(new_xptr_p(x))


def from_xptr(s):
    return ffi.from_handle(lib.R_ExternalPtrAddr(unbox(s)))



from .setup import init, loop
from .callbacks import def_callback, undef_callback
from .interface import rparse, reval, rprint, rlang, rcall, rcopy, robject


__all__ = [
    "init",
    "loop",
    "def_callback",
    "undef_callback",
    "rparse",
    "reval",
    "rprint",
    "rlang",
    "rcall",
    "rcopy",
    "robject"
]

__version__ = '0.3.30'



# a simple repl loop suitable used within python console
from __future__ import unicode_literals

from prompt_toolkit.shortcuts import prompt

from .interface import rprint, rcopy, reval, rcall
from .interface import process_events


__all__ = ["repl_r"]


def inputhook(context):
    while True:
        if context.input_is_ready():
            break
        process_events()


def repl_r():
    while True:
        try:
            text = prompt("> ", inputhook=inputhook)
            result = reval("withVisible({{{}}})".format(text))
            if rcopy(rcall("$", result, "visible")):
                rprint(rcall("$", result, "value"))
        except RuntimeError:
            pass
        except EOFError:
            break
        except KeyboardInterrupt:
            pass
        except Exception as e:
            raise e



from __future__ import unicode_literals
from rchitect._cffi import ffi, lib
from .utils import rconsole2str, utf8tosystem
from . import console


class Callback(object):
    suicide = None
    show_message = None
    read_console = None
    write_console_ex = None
    reset_console = None
    flush_console = None
    clearerr_console = None
    busy = None
    clean_up = None
    show_files = None
    choose_file = None
    edit_file = None
    loadhistory = None
    savehistory = None
    addhistory = None
    edit_files = None
    do_selectlist = None
    do_dataentry = None
    do_dataviewer = None
    process_events = None
    polled_events = None
    yes_no_cancel = None

    def __setattr__(self, item, value):
        if item not in Callback.__dict__.keys():
            raise KeyError()
        self.__dict__[item] = value


callback = Callback()
console.reg_callback(callback)


def def_callback(name=None):
    def _(fun):
        fname = name
        if fname is None:
            fname = fun.__name__
        setattr(callback, fname, fun)

    return _


def undef_callback(name):
    setattr(callback, name, None)


# prevent rstart being gc'ed
_protected = {}


def setup_rstart(rhome, args):
    rstart = ffi.new("Rstart")
    _protected["rstart"] = rstart
    SA_NORESTORE = 0
    SA_RESTORE = 1
    # SA_DEFAULT = 2
    SA_NOSAVE = 3
    SA_SAVE = 4
    SA_SAVEASK = 5
    # SA_SUICIDE = 6
    lib.R_DefParams(rstart)
    rstart.R_Quiet = "--quiet" in args
    rstart.R_Slave = "--slave" in args
    rstart.R_Interactive = 1
    rstart.R_Verbose = "--verbose" in args
    rstart.LoadSiteFile = "--no-site-file" not in args
    rstart.LoadInitFile = "--no-init-file" not in args
    if "--no-restore" in args:
        rstart.RestoreAction = SA_NORESTORE
    else:
        rstart.RestoreAction = SA_RESTORE
    if "--no-save" in args:
        rstart.SaveAction = SA_NOSAVE
    elif "--save" in args:
        rstart.SaveAction = SA_SAVE
    else:
        rstart.SaveAction = SA_SAVEASK
    rhome = ffi.new("char[]", rhome.encode("utf-8"))
    _protected["rhome"] = rhome
    rstart.rhome = rhome
    home = ffi.new("char[]", ffi.string(lib.getRUser()))
    _protected["home"] = home
    rstart.home = home
    rstart._ReadConsole = ffi.addressof(lib, "cb_read_console_interruptible")
    rstart._WriteConsole = ffi.NULL
    rstart.CallBack = ffi.addressof(lib, "cb_polled_events_interruptible")
    rstart.ShowMessage = ffi.addressof(lib, "cb_show_message")
    rstart.YesNoCancel = ffi.addressof(lib, "cb_yes_no_cancel")
    rstart.Busy = ffi.addressof(lib, "cb_busy")
    # we cannot get it to RGui, otherwise `do_system` will clear the standard handlers
    rstart.CharacterMode = 1  # RTerm
    rstart.WriteConsoleEx = ffi.addressof(lib, "cb_write_console_capturable")
    lib.R_SetParams(rstart)


def setup_callback(p, name, cb_name=None):
    if name is None:
        lib._libR_set_callback(p.encode(), ffi.NULL)
    elif getattr(callback, name):
        cb_name = cb_name if cb_name is not None else "cb_" + name
        lib._libR_set_callback(p.encode(), ffi.addressof(lib, str(cb_name)))


def setup_unix_callbacks():
    setup_callback("R_Outputfile", None)
    setup_callback("R_Consolefile", None)

    setup_callback("ptr_R_Suicide", "suicide")
    setup_callback("ptr_R_ShowMessage", "show_message")
    setup_callback("ptr_R_ReadConsole", "read_console", "cb_read_console_interruptible")
    setup_callback("ptr_R_WriteConsole", None)
    setup_callback("ptr_R_WriteConsoleEx", "write_console_ex", "cb_write_console_safe")
    setup_callback("ptr_R_ResetConsole", "reset_console")
    setup_callback("ptr_R_FlushConsole", "flush_console")
    setup_callback("ptr_R_ClearerrConsole", "clearerr_console")
    setup_callback("ptr_R_Busy", "busy", "cb_busy_safe")
    setup_callback("ptr_R_CleanUp", "clean_up")
    setup_callback("ptr_R_ShowFiles", "show_files")
    setup_callback("ptr_R_ChooseFile", "choose_file")
    setup_callback("ptr_R_EditFile", "edit_file")
    setup_callback("ptr_R_loadhistory", "loadhistory")
    setup_callback("ptr_R_savehistory", "savehistory")
    setup_callback("ptr_R_addhistory", "addhistory")
    setup_callback("ptr_R_EditFiles", "edit_files")
    setup_callback("ptr_do_selectlist", "do_selectlist")
    setup_callback("ptr_do_dataentry", "do_dataentry")
    setup_callback("ptr_do_dataviewer", "do_dataviewer")
    setup_callback("ptr_R_ProcessEvents", "process_events")
    setup_callback("R_PolledEvents", "polled_events", "cb_polled_events_interruptible")


@ffi.def_extern()
def cb_show_message(buf):
    callback.show_message(rconsole2str(ffi.string(buf)))


def on_callback_error(exception, exc_value, traceback):
    if exception == KeyboardInterrupt:
        lib.cb_interrupted = 1
    elif exception == EOFError:
        pass
    else:
        print(exception, exc_value)


_code = [""]


@ffi.def_extern(error=0, onerror=on_callback_error)
def cb_read_console(p, buf, buflen, add_history):
    # cache the code as buflen is limited to 4096
    if _code[0]:
        code = _code[0]
    else:
        text = callback.read_console(rconsole2str(ffi.string(p)), add_history)
        if text is None:
            return 0
        code = utf8tosystem(text)
        _code[0] = code

    buf = ffi.cast("char*", buf)
    nb = min(len(code), buflen - 2)
    buf[0:nb] = code[0:nb]
    if nb < buflen - 2:
        buf[nb] = b'\n'
        buf[nb + 1] = b'\x00'

    _code[0] = _code[0][nb:]
    return 1


@ffi.def_extern()
def cb_write_console_capturable(buf, bufline, otype):
    text = rconsole2str(ffi.string(buf))
    console.write_console(text, otype)


@ffi.def_extern()
def cb_busy(which):
    callback.busy(which)


@ffi.def_extern()
def cb_clean_up(saveact, status, run_last):
    callback.clean_up(saveact, status, run_last)


@ffi.def_extern(error=None, onerror=on_callback_error)
def cb_polled_events():
    console.flush()
    callback.polled_events()


@ffi.def_extern()
def cb_yes_no_cancel(p):
    return callback.yes_no_cancel(rconsole2str(ffi.string(p)))



from __future__ import unicode_literals
from six import raise_from

# much of the following mechanism is inspired by https://github.com/mrocklin/multipledispatch


AMBIGUITY = "Signature {} is ambiguous with {}. Define signature {} to resolve the ambiguity"


def expand_tuples(r):
    """
    >>> expand_tuples([1, (2, 3)])
    [(1, 2), (1, 3)]
    >>> expand_tuples([1, 2])
    [(1, 2)]
    """
    if not r:
        return [()]
    elif not isinstance(r[0], tuple):
        rest = expand_tuples(r[1:])
        return [(r[0],) + t for t in rest]
    else:
        rest = expand_tuples(r[1:])
        return [(item,) + t for t in rest for item in r[0]]


def isstrictsubclass(a, b):
    return a != b and issubclass(a, b)


class Dispatcher(object):
    # share acrossed dispatchers
    _typeof_mapping = {}

    def __init__(self):
        self._cache = {}
        self._ordering = []
        self.funcs = {}

    @classmethod
    def add_dispatch_policy(cls, t, typeof):
        cls._typeof_mapping[t] = typeof

    def __call__(self, *args, **kwargs):
        types = tuple([self.typeof(arg) for arg in args])
        try:
            func = self._cache[types]
        except KeyError:
            func = self.dispatch(*types)
            if not func:
                raise_from(NotImplementedError("Dispatch not found for signature %s" % str(types)), None)
            self._cache[types] = func
        return func(*args, **kwargs)

    def typeof(self, arg):
        t = type(arg)
        try:
            return self._typeof_mapping[t](arg)
        except Exception:
            pass
        return t

    @property
    def ordering(self):
        return self._ordering

    def reorder(self, types):
        if types in self._ordering:
            raise TypeError("Signature %s redefined" % str(types))

        n = len(types)
        for signature in self._ordering:
            if len(signature) != n:
                continue
            if not any(map(isstrictsubclass, types, signature)) or \
                    not any(map(isstrictsubclass, signature, types)):
                continue
            parent = tuple(types[i] if issubclass(types[i], signature[i]) else signature[i]
                           for i in range(len(types)))
            if parent not in self._ordering:
                raise TypeError(
                    AMBIGUITY.format(
                        str(types),
                        str(signature),
                        str(parent)))

        p = 0
        for signature in self._ordering:
            if len(signature) == n and all(map(issubclass, types, signature)):
                break
            p = p + 1
        self._ordering.insert(p, types)

    def add(self, types, func):
        # support union types
        if any(isinstance(typ, tuple) for typ in types):
            for typs in expand_tuples(types):
                self.add(typs, func)
            return

        self.reorder(types)
        self._cache.clear()
        self.funcs[types] = func

    def dispatch(self, *types):
        if types in self.funcs:
            return self.funcs[types]
        try:
            return next(self.dispatch_iter(*types))
        except StopIteration:
            return None

    def dispatch_iter(self, *types):
        n = len(types)
        for signature in self.ordering:
            if len(signature) == n and all(map(issubclass, types, signature)):
                result = self.funcs[signature]
                yield result


namespace = dict()


def dispatch(*types):

    types = tuple(types)

    def _(func):
        name = func.__name__
        if name not in namespace:
            namespace[name] = Dispatcher()

        dispatcher = namespace[name]
        dispatcher.add(types, func)
        return dispatcher
    return _


dispatch.add_dispatch_policy = Dispatcher.add_dispatch_policy



from __future__ import unicode_literals, absolute_import
from rchitect._cffi import lib

import operator
import sys
import importlib
from six import text_type
from types import ModuleType

from .interface import rcopy, robject, rfunction, rcall_p, rcall


def get_p(name, envir):
    return rcall_p(("base", "get"), name, envir=envir)


def inject_py_tools():

    def py_import(module):
        return importlib.import_module(module)

    def py_import_builtins():
        if sys.version >= "3":
            return importlib.import_module("builtins")
        else:
            return importlib.import_module("__builtin__")

    def py_call(fun, *args, **kwargs):
        # todo: suuport .asis and .convert
        if isinstance(fun, text_type):
            fun = eval(fun)
        return fun(*args, **kwargs)

    def py_copy(*args):
        return robject(*args)

    def py_eval(code):
        return eval(code)

    def py_get_attr(obj, key):
        if isinstance(obj, ModuleType):
            try:
                return importlib.import_module("{}.{}".format(obj.__name__, key))
            except ImportError:
                pass
        return getattr(obj, key)

    def py_get_item(obj, key):
        return obj[key]

    def py_names(obj):
        try:
            return list(k for k in obj.__dict__.keys() if not k.startswith("_"))
        except Exception:
            return None

    def py_object(*args):
        if len(args) == 1:
            return robject("PyObject", rcopy(args[0]))
        elif len(args) == 2:
            return robject("PyObject", rcopy(rcopy(object, args[0]), args[1]))

    def py_print(r):
        rcall_p("cat", repr(r) + "\n")

    def py_set_attr(obj, key, value):
        lib.Rf_protect(obj)
        lib.Rf_protect(key)
        lib.Rf_protect(value)
        pyo = rcopy(object, obj)
        try:
            setattr(pyo, rcopy(key), rcopy(value))
        finally:
            lib.Rf_unprotect(3)
        return pyo

    def py_set_item(obj, key, value):
        lib.Rf_protect(obj)
        lib.Rf_protect(key)
        lib.Rf_protect(value)
        pyo = rcopy(object, obj)
        try:
            pyo[rcopy(key)] = rcopy(value)
        finally:
            lib.Rf_unprotect(3)
        return pyo

    def py_dict(**kwargs):
        narg = len(kwargs)
        for key in kwargs:
            lib.Rf_protect(kwargs[key])
        try:
            return {key: rcopy(kwargs[key]) for key in kwargs}
        finally:
            lib.Rf_unprotect(narg)

    def py_tuple(*args):
        narg = len(args)
        for a in args:
            lib.Rf_protect(a)
        try:
            return tuple([rcopy(a) for a in args])
        finally:
            lib.Rf_unprotect(narg)

    def py_unicode(obj):
        return text_type(obj)

    def assign(name, value, envir):
        rcall(("base", "assign"), name, value, envir=envir)

    e = rcall(("base", "new.env"), parent=lib.R_GlobalEnv)
    kwarg = {"rchitect.py_tools": e}
    rcall(("base", "options"), **kwarg)

    assign("import", rfunction(py_import, convert=False), e)
    assign("import_builtins", rfunction(py_import_builtins, convert=False), e)
    assign("py_call", rfunction(py_call, convert=False), e)
    assign("py_copy", rfunction(py_copy, convert=True), e)
    assign("py_eval", rfunction(py_eval, convert=False), e)
    assign("py_get_attr", rfunction(py_get_attr, convert=False), e)
    assign("py_get_item", rfunction(py_get_item, convert=False), e)
    assign("py_object", rfunction(py_object, asis=True, convert=False), e)
    assign("py_set_attr", rfunction(py_set_attr, invisible=True, asis=True, convert=False), e)
    assign("py_set_item", rfunction(py_set_item, invisible=True, asis=True, convert=False), e)
    assign("py_unicode", rfunction(py_unicode, convert=False), e)
    assign("dict", rfunction(py_dict, asis=True, convert=False), e)
    assign("tuple", rfunction(py_tuple, asis=True, convert=False), e)

    assign("names.PyObject", rfunction(py_names, convert=True), e)
    assign("print.PyObject", rfunction(py_print, invisible=True, convert=False), e)
    assign(".DollarNames.PyObject", rfunction(py_names, convert=True), e)
    assign("$.PyObject", rfunction(py_get_attr, convert=True), e)
    assign("[.PyObject", rfunction(py_get_item, convert=True), e)
    assign("$<-.PyObject", rfunction(py_set_attr, invisible=True, asis=True, convert=False), e)
    assign("[<-.PyObject", rfunction(py_set_item, invisible=True, asis=True, convert=False), e)
    assign("&.PyObject", rfunction(operator.and_, invisible=True, convert=False), e)
    assign("|.PyObject", rfunction(operator.or_, invisible=True, convert=False), e)
    assign("!.PyObject", rfunction(operator.not_, invisible=True, convert=False), e)

    def attach():
        parent_frame = rcall("sys.frame", -1)
        things = [
            "import",
            "import_builtins",
            "py_call",
            "py_copy",
            "py_eval",
            "py_get_attr",
            "py_get_item",
            "py_object",
            "py_set_attr",
            "py_set_item",
            "py_unicode",
            "dict",
            "tuple",
            "names.PyObject",
            "print.PyObject",
            ".DollarNames.PyObject",
            "$.PyObject",
            "[.PyObject",
            "$<-.PyObject",
            "[<-.PyObject",
            "&.PyObject",
            "|.PyObject",
            "!.PyObject"
        ]
        for thing in things:
            assign(thing, get_p(thing, e), parent_frame)

    assign("attach", robject(attach, invisible=True), e)



from __future__ import unicode_literals
from rchitect._cffi import ffi, lib
# from future.utils import with_metaclass


_sexpnums = {
    "NILSXP": 0,
    "SYMSXP": 1,
    "LISTSXP": 2,
    "CLOSXP": 3,
    "ENVSXP": 4,
    "PROMSXP": 5,
    "LANGSXP": 6,
    "SPECIALSXP": 7,
    "BUILTINSXP": 8,
    "CHARSXP": 9,
    "LGLSXP": 10,
    "INTSXP": 13,
    "REALSXP": 14,
    "CPLXSXP": 15,
    "STRSXP": 16,
    "DOTSXP": 17,
    "ANYSXP": 18,
    "VECSXP": 19,
    "EXPRSXP": 20,
    "BCODESXP": 21,
    "EXTPTRSXP": 22,
    "WEAKREFSXP": 23,
    "RAWSXP": 24,
    "S4SXP": 25,
    "NEWSXP": 30,
    "FREESXP": 31,
    "FUNSXP": 99
}

_sexpnums_reversed = {v: k for (k, v) in _sexpnums.items()}


class sexptype(type):

    def __new__(cls, *args):
        if len(args) == 1:
            x = args[0]
            assert isinstance(x, ffi.CData)
            assert ffi.typeof(x) == ffi.typeof('SEXP')
            return globals()[_sexpnums_reversed[lib.TYPEOF(x)]]
        else:
            x = str(args[0])
            return super(sexptype, cls).__new__(cls, x, *args[1:])

    def __instancecheck__(self, instance):
        if not isinstance(instance, ffi.CData):
            return False
        if ffi.typeof(instance) != ffi.typeof('SEXP'):
            return False
        if self is SEXP:
            return True
        return _sexpnums[self.__name__] == lib.TYPEOF(instance)

    def __subclasscheck__(self, subclass):
        if subclass == ffi.CData:
            return True
        return super(sexptype, self).__subclasscheck__(subclass)

    def __repr__(self):
        return self.__name__


SEXP = sexptype("SEXP", (object, ), {})
NILSXP = sexptype("NILSXP", (SEXP, ), {})
SYMSXP = sexptype("SYMSXP", (SEXP, ), {})
LISTSXP = sexptype("LISTSXP", (SEXP, ), {})
CLOSXP = sexptype("CLOSXP", (SEXP, ), {})
ENVSXP = sexptype("ENVSXP", (SEXP, ), {})
PROMSXP = sexptype("PROMSXP", (SEXP, ), {})
LANGSXP = sexptype("LANGSXP", (SEXP, ), {})
SPECIALSXP = sexptype("SPECIALSXP", (SEXP, ), {})
BUILTINSXP = sexptype("BUILTINSXP", (SEXP, ), {})
CHARSXP = sexptype("CHARSXP", (SEXP, ), {})
LGLSXP = sexptype("LGLSXP", (SEXP, ), {})
INTSXP = sexptype("INTSXP", (SEXP, ), {})
REALSXP = sexptype("REALSXP", (SEXP, ), {})
CPLXSXP = sexptype("CPLXSXP", (SEXP, ), {})
STRSXP = sexptype("STRSXP", (SEXP, ), {})
DOTSXP = sexptype("DOTSXP", (SEXP, ), {})
ANYSXP = sexptype("ANYSXP", (SEXP, ), {})
VECSXP = sexptype("VECSXP", (SEXP, ), {})
EXPRSXP = sexptype("EXPRSXP", (SEXP, ), {})
BCODESXP = sexptype("BCODESXP", (SEXP, ), {})
EXTPTRSXP = sexptype("EXTPTRSXP", (SEXP, ), {})
WEAKREFSXP = sexptype("WEAKREFSXP", (SEXP, ), {})
RAWSXP = sexptype("RAWSXP", (SEXP, ), {})
S4SXP = sexptype("S4SXP", (SEXP, ), {})
NEWSXP = sexptype("NEWSXP", (SEXP, ), {})
FREESXP = sexptype("FREESXP", (SEXP, ), {})
FUNSXP = sexptype("FUNSXP", (SEXP, ), {})


class RObject(object):
    def __init__(self, s):
        assert isinstance(s, SEXP)
        self.s = s
        lib.R_PreserveObject(self.s)

    def __del__(self):
        try:
            # it might cause AttributeError when the program exits
            lib.R_ReleaseObject(self.s)
        except AttributeError:
            pass


class RClass(object):
    _rclasses = {}

    def __new__(cls, rcls):
        try:
            return cls._rclasses[rcls]
        except KeyError:
            T = type(str("RClass_" + rcls), (RClass,), {"__new__": (lambda cls: None)})
            cls._rclasses[rcls] = T
            return T


class datatype(type):
    """
    We introduce a `datatype` function to allow different values from `datatype(str)` and
    `datatype(int)`.
    """

    _datatypes = {}

    def __new__(cls, *args):
        if len(args) == 1:
            t = args[0]
            try:
                return cls._datatypes[t]
            except KeyError:
                T = datatype(
                    str("datatype_{}".format(t.__name__)),
                    (type,),
                    {"t": t, "__new__": (lambda cls: cls.t)})
                cls._datatypes[t] = T
                return T
        else:
            return super(datatype, cls).__new__(cls, *args)

    def __instancecheck__(self, instance):
        return self.t == instance

    def __subclasscheck__(self, subclass):
        return isinstance(subclass, datatype) and self.t == subclass.t

    def __repr__(self):
        return "datatype({})".format(self.t.__name__)


def box(x):
    if isinstance(x, SEXP):
        return RObject(x)
    elif isinstance(x, RObject):
        return x

    raise TypeError("expect SEXP or RObject")


def unbox(x):
    if isinstance(x, RObject):
        return x.s
    elif isinstance(x, SEXP):
        return x

    raise TypeError("expect SEXP or RObject")



from __future__ import unicode_literals
import sys
import os
from six.moves import input as six_input

from rchitect._cffi import ffi, lib
from .utils import Rhome, ensure_path, system2utf8
from .callbacks import def_callback, setup_unix_callbacks, setup_rstart


def load_lib_error():
    return"Cannot load shared library: {}".format(
        system2utf8(ffi.string(lib._libR_dl_error_message())))


def load_symbol_error():
    return "Cannot load symbol {}: {}".format(
                system2utf8(ffi.string(lib._libR_last_loaded_symbol())),
                system2utf8(ffi.string(lib._libR_dl_error_message())))


def load_constant_error():
    return "Cannot load constant {}: {}".format(
                system2utf8(ffi.string(lib._libR_last_loaded_symbol())),
                system2utf8(ffi.string(lib._libR_dl_error_message())))


def init(args=None, register_signal_handlers=None):

    if not args:
        args = ["rchitect", "--quiet", "--no-save"]

    if register_signal_handlers is None:
        register_signal_handlers = os.environ.get("RCHITECT_REGISTER_SIGNAL_HANDLERS", "1") == "1"

    rhome = Rhome()
    # microsoft python doesn't load DLL's from PATH
    # we will need to open the DLL's directly in _libR_load
    ensure_path(rhome)

    libR_loaded = lib.Rf_initialize_R != ffi.NULL

    if not libR_loaded:
        # `system2utf8` may not work before `Rf_initialize_R` because locale may not be set
        if not lib._libR_load(rhome.encode("utf-8")):
            raise Exception(load_lib_error())
        if not lib._libR_load_symbols():
            raise Exception(load_symbol_error())

    # _libR_is_initialized only works after _libR_load is run.
    if not lib._libR_is_initialized():

        _argv = [ffi.new("char[]", a.encode("utf-8")) for a in args]
        argv = ffi.new("char *[]", _argv)

        if sys.platform.startswith("win"):
            if register_signal_handlers:
                lib.Rf_initialize_R(len(argv), argv)
                setup_rstart(rhome, args)
            else:
                # Rf_initialize_R will set handler for SIGINT
                # we need to workaround it
                lib.R_SignalHandlers_t[0] = 0
                setup_rstart(rhome, args)
                lib.R_set_command_line_arguments(len(argv), argv)
                lib.GA_initapp(0, ffi.NULL)
            lib.setup_Rmainloop()
            # require R 4.0
            if lib.EmitEmbeddedUTF8_t != ffi.NULL:
                lib.EmitEmbeddedUTF8_t[0] = 1
        else:
            lib.R_SignalHandlers_t[0] = register_signal_handlers
            lib.Rf_initialize_R(len(argv), argv)
            setup_unix_callbacks()
            lib.setup_Rmainloop()

    if not libR_loaded:
        if not lib._libR_load_constants():
            raise Exception(load_constant_error())
        lib._libR_setup_xptr_callback()

        from rchitect.py_tools import inject_py_tools
        inject_py_tools()

        if os.environ.get("RCHITECT_RETICULATE_CONFIG", "1") != "0":
            from rchitect import reticulate
            reticulate.configure()


def loop():
    lib.run_Rmainloop()


def ask_input(s):
    return six_input(s)


@def_callback()
def show_message(buf):
    sys.stdout.write(buf)
    sys.stdout.flush()


@def_callback()
def read_console(p, add_history):
    sys.stdout.flush()
    sys.stderr.flush()
    return ask_input(p)


@def_callback()
def write_console_ex(buf, otype):
    if otype == 0:
        if sys.stdout:
            sys.stdout.write(buf)
            sys.stdout.flush()
    else:
        if sys.stderr:
            sys.stderr.write(buf)
            sys.stderr.flush()


@def_callback()
def busy(which):
    pass


@def_callback()
def polled_events():
    pass


# @def_callback()
# def clean_up(saveact, status, run_last):
#     lib.Rstd_CleanUp(saveact, status, run_last)


@def_callback()
def yes_no_cancel(p):
    while True:
        try:
            result = ask_input("{} [y/n/c]: ".format(p))
            if result in ["Y", "y"]:
                return 1
            elif result in ["N", "n"]:
                return 2
            else:
                return 0
        except EOFError:
            return 0
        except KeyboardInterrupt:
            return 0
        except Exception:
            pass



from __future__ import unicode_literals

from rchitect.interface import rcall, reval, rcopy, rstring, rint
from six import text_type


def assign_line_buffer(buf):
    rcall(reval("utils:::.assignLinebuffer"), rstring(buf))
    rcall(reval("utils:::.assignEnd"), rint(len(buf)))
    token = rcopy(text_type, rcall(reval("utils:::.guessTokenFromLine")))
    return token


CompleteCode = """
tryCatch(
    {{
        if ({settimelimit}) base::setTimeLimit({timeout})
        utils:::.completeToken()
        if ({settimelimit}) base::setTimeLimit()
    }},
    error = function(e) {{
        if ({settimelimit}) base::setTimeLimit()
        assign("comps", NULL, envir = utils:::.CompletionEnv)
    }}
)
"""


def complete_token(timeout=0):
    try:
        reval(
            CompleteCode.format(
                settimelimit="TRUE" if timeout > 0 else "FALSE",
                timeout=str(timeout)),
            envir=rcall("new.env")
        )
    except Exception:
        if timeout:
            rcall(("base", "setTimeLimit"))
            rcall(("base", "assign"), "comps", None, env=reval("utils:::.CompletionEnv"))


def retrieve_completions():
    completions = rcopy(list, rcall(reval("utils:::.retrieveCompletions")))
    if not completions:
        return []
    else:
        return completions



from __future__ import unicode_literals, absolute_import
from rchitect._cffi import ffi, lib
from .dispatch import dispatch
from .types import RObject, SEXP, RClass, sexptype, datatype
from .types import NILSXP, CLOSXP, ENVSXP, BUILTINSXP, LGLSXP, INTSXP, REALSXP, CPLXSXP, STRSXP, \
    VECSXP, EXPRSXP, EXTPTRSXP, RAWSXP, S4SXP
from .types import box, unbox
from .xptr import new_xptr, new_xptr_p, from_xptr
from .console import capture_console, read_stdout, read_stderr
from .utils import utf8tosystem


import sys
import ctypes
import struct
from contextlib import contextmanager
from six import text_type, string_types
from types import FunctionType
from collections import OrderedDict
if sys.version >= "3":
    from collections.abc import Callable
    long = int
else:
    from collections import Callable

dispatch.add_dispatch_policy(type, datatype)
dispatch.add_dispatch_policy(
    ffi.CData,
    lambda x: sexptype(x) if ffi.typeof(x) == ffi.typeof('SEXP') else ffi.CData)


def extract(kwargs, key, default=None):
    if key in kwargs:
        value = kwargs[key]
        del kwargs[key]
    else:
        value = default
    return value


def ensure_initialized():
    if lib.Rf_initialize_R == ffi.NULL:
        from .setup import init
        init()


@contextmanager
def protected(*args):
    nprotect = 0
    for a in args:
        if isinstance(a, SEXP):
            lib.Rf_protect(a)
            nprotect += 1
    try:
        yield
    finally:
        lib.Rf_unprotect(nprotect)


def rint_p(s):
    return lib.Rf_ScalarInteger(s)


def rint(s):
    return box(rint_p(s))


def rlogical_p(s):
    return lib.Rf_ScalarLogical(s)


def rlogical(s):
    return box(rlogical_p(s))


def rdouble_p(s):
    return lib.Rf_ScalarReal(s)


def rdouble(s):
    return box(rdouble_p(s))


def rchar_p(s):
    isascii = all(ord(c) < 128 for c in s)
    b = s.encode('utf-8')
    return lib.Rf_mkCharLenCE(b, len(b), lib.CE_NATIVE if isascii else lib.CE_UTF8)


def rchar(s):
    return box(rchar_p(s))


def rstring_p(s):
    return lib.Rf_ScalarString(rchar_p(s))


def rstring(s):
    return box(rstring_p(s))


def rsym_p(s, t=None):
    if t:
        return rlang_p("::", rsym_p(s), rsym_p(t))
    else:
        return lib.Rf_install(utf8tosystem(s))


def rsym(s, t=None):
    return box(rsym_p(s, t))


def rparse_p(s):
    ensure_initialized()
    status = ffi.new("ParseStatus[1]")
    s = lib.Rf_mkString(utf8tosystem(s))
    with protected(s):
        with capture_console():
            ret = lib.R_ParseVector(s, -1, status, lib.R_NilValue)
            if status[0] != lib.PARSE_OK:
                err = read_stderr().strip() or "Error"
                raise RuntimeError("{}".format(err))
    return ret


def rparse(s):
    return box(rparse_p(s))


@dispatch(EXPRSXP)
def reval_p(s, envir=None):
    ensure_initialized()
    if envir:
        # `sys.frame()` doesn't work with R_tryEval as it doesn't create R stacks,
        # we use `base::eval` instead.
        ret = rcall_p(("base", "eval"), s, _envir=envir)
    else:
        ret = lib.R_NilValue
        status = ffi.new("int[1]")
        with protected(s):
            with capture_console():
                for i in range(0, lib.Rf_length(s)):
                    ret = lib.R_tryEval(lib.VECTOR_ELT(s, i), lib.R_GlobalEnv, status)
                    if status[0] != 0:
                        err = read_stderr().strip() or "Error"
                        raise RuntimeError("{}".format(err))
    return ret


@dispatch(RObject)  # noqa
def reval_p(s, envir=None):
    with protected(envir):
        return reval_p(unbox(s), envir=envir)


@dispatch(string_types)  # noqa
def reval_p(s, envir=None):
    with protected(envir):
        return reval_p(rparse_p(s), envir=envir)


def reval(s, envir=None):
    return box(reval_p(s, envir=envir))


def as_call(x):
    if isinstance(x, SEXP):
        return x
    elif isinstance(x, RObject):
        return unbox(x)
    elif isinstance(x, string_types):
        return rsym_p(x)
    elif isinstance(x, tuple) and len(x) == 2:
        return rsym_p(*x)

    raise TypeError("unexpected function")


def rlang_p(f, *args, **kwargs):
    ensure_initialized()
    argslist = list(args) + list(kwargs.items())
    with protected(*argslist):
        nargs = len(args) + len(kwargs)
        t = lib.Rf_allocVector(lib.LANGSXP, nargs + 1)

        with protected(t):
            s = t
            lib.SETCAR(s, as_call(f))

            for a in args:
                s = lib.CDR(s)
                lib.SETCAR(s, unbox(a))
            for k, v in kwargs.items():
                s = lib.CDR(s)
                lib.SETCAR(s, unbox(v))
                lib.SET_TAG(s, lib.Rf_install(utf8tosystem(k)))

            ret = t
    return ret


def rlang(*args, **kwargs):
    return box(rlang_p(*args, **kwargs))


@contextmanager
def sexp_args(args, kwargs, asis=False):
    nprotect = 0
    try:
        _args = []
        for a in args:
            _a = sexp_as_py_object(a) if asis else sexp(a)
            if isinstance(_a, SEXP):
                lib.Rf_protect(_a)
                nprotect += 1
            _args.append(_a)

        _kwargs = {}
        for k, v in kwargs.items():
            _v = sexp_as_py_object(v) if asis else sexp(v)
            if isinstance(_v, SEXP):
                lib.Rf_protect(_v)
                nprotect += 1
            _kwargs[k] = _v

        yield _args, _kwargs
    finally:
        lib.Rf_unprotect(nprotect)


def rcall_p(f, *args, **kwargs):
    ensure_initialized()
    _envir = extract(kwargs, "_envir")
    _asis = extract(kwargs, "_asis")
    if _envir:
        _envir = unbox(_envir)
    else:
        _envir = lib.R_GlobalEnv

    with protected(_envir):
        with sexp_args(args, kwargs, _asis) as (a, k):
            with capture_console():
                status = ffi.new("int[1]")
                val = lib.R_tryEval(rlang_p(f, *a, **k), _envir, status)
                if status[0] != 0:
                    err = read_stderr().strip() or "Error"
                    raise RuntimeError("{}".format(err))
    return val


def rcall(*args, **kwargs):
    _convert = extract(kwargs, "_convert")
    s = rcall_p(*args, **kwargs)
    return rcopy(s) if _convert else box(s)


def rprint(s):
    new_env = rcall(("base", "new.env"))
    lib.Rf_defineVar(rsym_p("x"), unbox(s), unbox(new_env))
    with protected(s):
        try:
            rcall(("base", "print"), rsym("x"), _envir=new_env)
        finally:
            lib.Rf_defineVar(rsym_p("x"), lib.R_NilValue, unbox(new_env))


def _repr(self):
    s = self.s
    new_env = rcall(("base", "new.env"))
    lib.Rf_defineVar(rsym_p("x"), unbox(s), unbox(new_env))
    with capture_console(flushable=False):
        try:
            rcall(("base", "print"), rsym_p("x"), _envir=new_env)
        finally:
            lib.Rf_defineVar(rsym_p("x"), lib.R_NilValue, unbox(new_env))
        output = read_stdout() or ""

    name = "RObject{{{}}}".format(str(sexptype(s)))
    if output:
        return name + "\n" + output.rstrip()
    else:
        return name


RObject.__repr__ = _repr


def getoption_p(key):
    return lib.Rf_GetOption1(rsym_p(key))


def getoption(key):
    return box(getoption_p(key))


def roption(key, default=None):
    ret = rcopy(getoption_p(key))
    return ret if ret is not None else default


def setoption(key, value):
    kwargs = {key: value}
    rcall_p(("base", "options"), **kwargs)


def getattrib_p(s, key):
    return lib.Rf_getAttrib(unbox(s), rsym_p(key) if isinstance(key, string_types) else key)


def setattrib(s, key, value):
    with protected(s, value):
        lib.Rf_setAttrib(unbox(s), rsym_p(key) if isinstance(key, string_types) else key, value)


def getnames_p(s):
    return getattrib_p(s, lib.R_NamesSymbol)


def rnames(s):
    return rcopy(list, getnames_p(s))


def getclass_p(s, singleString):
    return lib.R_data_class(unbox(s), singleString)


def setclass(s, classes):
    with protected(s):
        setattrib(s, lib.R_ClassSymbol, sexp(RClass("character"), classes))


def rclass(s, singleString=0):
    return rcopy(text_type if singleString else list, getclass_p(s, singleString))


def process_events():
    lib.process_events()


def polled_events():
    lib.polled_events()


def peek_event():
    return lib.peek_event()


def set_hook(event, fun):
    rcall(("base", "setHook"), event, fun)


def package_event(pkg, event):
    return rcall(("base", "packageEvent"), pkg, event)


def greeting():
    info = rcopy(rcall("R.Version"))
    return "{} -- \"{}\"\nPlatform: {} ({}-bit)\n".format(
        info["version.string"], info["nickname"], info["platform"], 8 * struct.calcsize("P"))


# R to Py


@dispatch(datatype(type(None)), NILSXP)  # noqa
def rcopy(_, s):
    return None


@dispatch(datatype(list), NILSXP)  # noqa
def rcopy(_, s):
    return []


@dispatch(datatype(int), INTSXP)  # noqa
def rcopy(_, s):
    return lib.INTEGER(s)[0]


@dispatch(datatype(list), INTSXP)  # noqa
def rcopy(_, s):
    return [lib.INTEGER(s)[i] for i in range(lib.Rf_length(s))]


@dispatch(datatype(bool), LGLSXP)  # noqa
def rcopy(_, s):
    return bool(lib.LOGICAL(s)[0])


@dispatch(datatype(list), LGLSXP)  # noqa
def rcopy(_, s):
    return [bool(lib.LOGICAL(s)[i]) for i in range(lib.Rf_length(s))]


@dispatch(datatype(float), REALSXP)  # noqa
def rcopy(_, s):
    return lib.REAL(s)[0]


@dispatch(datatype(list), REALSXP)  # noqa
def rcopy(_, s):
    return [lib.REAL(s)[i] for i in range(lib.Rf_length(s))]


@dispatch(datatype(complex), CPLXSXP)  # noqa
def rcopy(_, s):
    z = lib.COMPLEX(s)[0]
    return complex(z.r, z.i)


@dispatch(datatype(list), CPLXSXP)  # noqa
def rcopy(_, s):
    return [complex(lib.COMPLEX(s)[i].r, lib.COMPLEX(s)[i].i) for i in range(lib.Rf_length(s))]


def _string(s):
    return text_type(ffi.string(lib.Rf_translateCharUTF8(s)).decode("utf-8"))


@dispatch(datatype(bytes), RAWSXP)  # noqa
def rcopy(_, s):
    return ffi.string(lib.RAW(s), lib.Rf_length(s))


@dispatch(datatype(text_type), STRSXP)  # noqa
def rcopy(_, s):
    return _string(lib.STRING_ELT(s, 0))


@dispatch(datatype(list), STRSXP)  # noqa
def rcopy(_, s):
    return [_string(lib.STRING_ELT(s, i)) for i in range(lib.Rf_length(s))]


@dispatch(datatype(list), VECSXP)  # noqa
def rcopy(_, s):
    return [rcopy(lib.VECTOR_ELT(s, i)) for i in range(lib.Rf_length(s))]


@dispatch(datatype(tuple), VECSXP)  # noqa
def rcopy(_, s):
    return tuple(rcopy(list, s))


@dispatch(datatype(dict), VECSXP)  # noqa
def rcopy(_, s):
    ret = dict()
    names = rnames(s)
    for i in range(lib.Rf_length(s)):
        ret[names[i]] = rcopy(lib.VECTOR_ELT(s, i))
    return ret


@dispatch(datatype(OrderedDict), VECSXP)  # noqa
def rcopy(_, s):
    ret = OrderedDict()
    names = rnames(s)
    for i in range(lib.Rf_length(s)):
        ret[names[i]] = rcopy(lib.VECTOR_ELT(s, i))
    return ret


@dispatch(datatype(FunctionType), (CLOSXP, BUILTINSXP))  # noqa
def rcopy(_, s, envir=None, asis=False, convert=True):
    r = RObject(s)  # preserve the closure

    def _(*args, **kwargs):
        return rcall(r, *args, _envir=envir, _asis=asis, _convert=convert, **kwargs)

    return _


# PyObject
@dispatch(datatype(object), EXTPTRSXP)  # noqa
def rcopy(_, s):
    return from_xptr(s)


# reticulate's python.builtin.object
@dispatch(datatype(object), ENVSXP)  # noqa
def rcopy(_, s):
    x = rcall(("base", "get"), "pyobj", s)
    p = ffi.cast("uintptr_t", lib.R_ExternalPtrAddr(unbox(x)))
    d = long(p)
    obj = ctypes.cast(d, ctypes.py_object)
    return obj.value


# PyCallable or reticulate's python.builtin.function
@dispatch(datatype(object), CLOSXP)  # noqa
def rcopy(_, s):
    return rcopy(getattrib_p(s, "py_object"))


@dispatch(datatype(RObject), SEXP)  # noqa
def rcopy(_, s):
    return box(s)


@dispatch(datatype(RObject), RObject)  # noqa
def rcopy(_, s):
    return s


@dispatch(object, RObject)  # noqa
def rcopy(t, r, **kwargs):
    return rcopy(t, unbox(r), **kwargs)


# default conversions
default = RClass("default")


@dispatch(SEXP)  # noqa
def rcopy(s, **kwargs):
    ensure_initialized()
    for cls in rclass(s):
        T = rcopytype(RClass(cls), s)
        if T is not RObject:
            return rcopy(T, s, **kwargs)
    T = rcopytype(default, s)
    return rcopy(T, s, **kwargs)


@dispatch(RObject)  # noqa
def rcopy(r, **kwargs):
    return rcopy(unbox(r), **kwargs)


# PyObject
@dispatch(datatype(RClass("PyObject")), EXTPTRSXP)  # noqa
def rcopytype(_, s):
    return object


# PyCallable
@dispatch(datatype(RClass("PyCallable")), CLOSXP)  # noqa
def rcopytype(_, s):
    return object


# reticulate class
@dispatch(datatype(RClass("python.builtin.object")), ENVSXP)  # noqa
def rcopytype(_, s):
    return object


# reticulate function
@dispatch(datatype(RClass("python.builtin.function")), CLOSXP)  # noqa
def rcopytype(_, s):
    return object


@dispatch(datatype(default), NILSXP)  # noqa
def rcopytype(_, s):
    return type(None)


@dispatch(datatype(default), INTSXP)  # noqa
def rcopytype(_, s):
    return int if lib.Rf_length(s) == 1 else list


@dispatch(datatype(default), LGLSXP)  # noqa
def rcopytype(_, s):
    return bool if lib.Rf_length(s) == 1 else list


@dispatch(datatype(default), REALSXP)  # noqa
def rcopytype(_, s):
    return float if lib.Rf_length(s) == 1 else list


@dispatch(datatype(default), CPLXSXP)  # noqa
def rcopytype(_, s):
    return complex if lib.Rf_length(s) == 1 else list


@dispatch(datatype(default), RAWSXP)  # noqa
def rcopytype(_, s):
    return bytes


@dispatch(datatype(default), STRSXP)  # noqa
def rcopytype(_, s):
    return text_type if lib.Rf_length(s) == 1 else list


@dispatch(datatype(default), VECSXP)  # noqa
def rcopytype(_, s):
    return list if lib.Rf_isNull(getnames_p(s)) else OrderedDict


@dispatch(datatype(default), BUILTINSXP)  # noqa
def rcopytype(_, s):
    return FunctionType


@dispatch(datatype(default), CLOSXP)  # noqa
def rcopytype(_, s):
    return FunctionType


@dispatch(datatype(default), EXTPTRSXP)  # noqa
def rcopytype(_, s):
    return RObject


@dispatch(object, SEXP)  # noqa
def rcopytype(_, s):
    return RObject


# Py to R


# python to r conversions

def robject(*args, **kwargs):
    ensure_initialized()
    if len(args) == 2 and isinstance(args[0], string_types):
        return RObject(sexp(RClass(args[0]), args[1], **kwargs))
    elif len(args) == 1:
        return RObject(sexp(args[0], **kwargs))
    else:
        raise TypeError("wrong number of arguments or argument types")


def rfunction(x, **kwargs):
    return robject("function", x, **kwargs)


@dispatch(datatype(RClass("NULL")), type(None))  # noqa
def sexp(_, n):
    return lib.R_NilValue


@dispatch(datatype(RClass("logical")), bool)  # noqa
def sexp(_, s):
    return rlogical_p(s)


@dispatch(datatype(RClass("integer")), int)  # noqa
def sexp(_, s):
    return rint_p(s)


@dispatch(datatype(RClass("numeric")), float)  # noqa
def sexp(_, s):
    return rdouble_p(s)


@dispatch(datatype(RClass("complex")), complex)  # noqa
def sexp(_, s):
    c = ffi.new("Rcomplex*")
    c.r = s.real
    c.i = s.imag
    return lib.Rf_ScalarComplex(c[0])


@dispatch(datatype(RClass("character")), string_types)  # noqa
def sexp(_, s):
    return rstring_p(s)


if sys.version >= "3":
    @dispatch(datatype(RClass("raw")), bytes)  # noqa
    def sexp(_, s):
        n = len(s)
        x = lib.Rf_allocVector(lib.RAWSXP, n)
        with protected(x):
            p = lib.RAW(x)
            for i in range(n):
                p[i] = s[i]
        return x
else:
    @dispatch(datatype(RClass("raw")), bytes)  # noqa
    def sexp(_, s):
        n = len(s)
        x = lib.Rf_allocVector(lib.RAWSXP, n)
        with protected(x):
            p = lib.RAW(x)
            for i in range(n):
                p[i] = ord(s[i])
        return x


@dispatch(datatype(RClass("logical")), list)  # noqa
def sexp(_, s):
    n = len(s)
    x = lib.Rf_allocVector(lib.LGLSXP, n)
    with protected(x):
        p = lib.LOGICAL(x)
        for i in range(n):
            p[i] = s[i]
    return x


@dispatch(datatype(RClass("integer")), list)  # noqa
def sexp(_, s):
    n = len(s)
    x = lib.Rf_allocVector(lib.INTSXP, n)
    with protected(x):
        p = lib.INTEGER(x)
        for i in range(n):
            p[i] = s[i]
    return x


@dispatch(datatype(RClass("numeric")), list)  # noqa
def sexp(_, s):
    n = len(s)
    x = lib.Rf_allocVector(lib.REALSXP, n)
    with protected(x):
        p = lib.REAL(x)
        for i in range(n):
            p[i] = s[i]
    return x


@dispatch(datatype(RClass("complex")), list)  # noqa
def sexp(_, s):
    n = len(s)
    x = lib.Rf_allocVector(lib.CPLXSXP, n)
    with protected(x):
        p = lib.COMPLEX(x)
        for i in range(n):
            p[i].r = s[i].real
            p[i].i = s[i].imag
    return x


@dispatch(datatype(RClass("character")), list)  # noqa
def sexp(_, s):
    n = len(s)
    x = lib.Rf_allocVector(lib.STRSXP, n)
    with protected(x):
        for i in range(n):
            isascii = all(ord(c) < 128 for c in s[i])
            b = s[i].encode('utf-8')
            lib.SET_STRING_ELT(x, i, lib.Rf_mkCharLenCE(b, len(b), 0 if isascii else 1))
    return x



@dispatch(datatype(RClass("list")), (list, tuple))  # noqa
def sexp(_, s):
    n = len(s)
    x = lib.Rf_allocVector(lib.VECSXP, n)
    with protected(x):
        for i in range(n):
            lib.SET_VECTOR_ELT(x, i, sexp(s[i]))
    return x


@dispatch(datatype(RClass("list")), (dict, OrderedDict))  # noqa
def sexp(_, s):
    v = sexp(RClass("list"), list(s.values()))
    with protected(v):
        k = sexp(RClass("character"), list(s.keys()))
        with protected(k):
            lib.Rf_setAttrib(v, lib.R_NamesSymbol, k)
    return v


def sexp_dots():  # noqa
    s = lib.Rf_list1(lib.R_MissingArg)
    with protected(s):
        lib.SET_TAG(s, lib.R_DotsSymbol)
    return s


def sexp_as_py_object(obj):
    if callable(obj):
        return sexp(RClass("PyCallable"), obj)
    else:
        return sexp(RClass("PyObject"), obj)


def on_xptr_callback_error(exception, exc_value, traceback):
    lib.xptr_callback_error_occured = 1
    lib.xptr_callback_error_message = utf8tosystem(str(exc_value)[0:100])


@ffi.def_extern(error=ffi.NULL, onerror=on_xptr_callback_error)
def xptr_callback(exptr, arglist, asis, convert):
    asis = rcopy(bool, asis)
    convert = rcopy(bool, convert)
    f = from_xptr(exptr)
    args = []
    kwargs = {}
    names = rnames(arglist)
    for i in range(lib.Rf_length(arglist)):
        if asis:
            if names and names[i]:
                kwargs[names[i]] = lib.VECTOR_ELT(arglist, i)
            else:
                args.append(lib.VECTOR_ELT(arglist, i))
        else:
            if names and names[i]:
                kwargs[names[i]] = rcopy(lib.VECTOR_ELT(arglist, i))
            else:
                args.append(rcopy(lib.VECTOR_ELT(arglist, i)))

    ret = f(*args, **kwargs)
    return sexp(ret) if convert else sexp_as_py_object(ret)


@dispatch(datatype(RClass("function")), Callable)  # noqa
def sexp(_, f, invisible=False, asis=False, convert=True):
    fextptr = new_xptr(f)
    dotlist = rlang("list", lib.R_DotsSymbol)
    body = rlang(
        ".Call", rstring("_libR_xptr_callback"), fextptr, dotlist,
        rlogical(asis), rlogical(convert))
    if invisible:
        body = rlang("invisible", body)
    lang = rlang_p(rsym("function"), sexp_dots(), body)
    status = ffi.new("int[1]")
    val = lib.R_tryEval(lang, lib.R_GlobalEnv, status)
    return val


@dispatch(datatype(RClass("PyObject")), object)  # noqa
def sexp(_, s):
    if (isinstance(s, RObject) or isinstance(s, SEXP)) and "PyObject" in rclass(s):
        return unbox(s)
    p = new_xptr_p(s)
    setclass(p, "PyObject")
    return p


@dispatch(datatype(RClass("PyCallable")), Callable)  # noqa
def sexp(_, f, invisible=False, asis=False, convert=True):
    p = sexp(RClass("function"), f, invisible=invisible, asis=asis, convert=convert)
    setattrib(p, "py_object", sexp(RClass("PyObject"), f))
    setclass(p, ["PyCallable", "PyObject"])
    return p


# default conversions


@dispatch(type(None))  # noqa
def sexp(n):
    return lib.R_NilValue


@dispatch(SEXP)  # noqa
def sexp(x):
    return x


@dispatch(RObject)  # noqa
def sexp(r):
    return unbox(r)


@dispatch(object)  # noqa
def sexp(s, **kwargs):
    rcls = sexpclass(s)
    return sexp(RClass(rcls), s, **kwargs)


@dispatch(bool)  # noqa
def sexpclass(s):
    return "logical"


@dispatch(int)  # noqa
def sexpclass(s):
    return "integer"


@dispatch(float)  # noqa
def sexpclass(s):
    return "numeric"


@dispatch(complex)  # noqa
def sexpclass(s):
    return "complex"


@dispatch(string_types)  # noqa
def sexpclass(s):
    return "character"


@dispatch(list)  # noqa
def sexpclass(s):
    if all(isinstance(x, bool) for x in s):
        return "logical"
    elif all(isinstance(x, int) for x in s):
        return "integer"
    elif all(isinstance(x, float) for x in s):
        return "numeric"
    elif all(isinstance(x, complex) for x in s):
        return "complex"
    elif all(isinstance(x, string_types) for x in s):
        return "character"

    return "list"


@dispatch((tuple, dict, OrderedDict))  # noqa
def sexpclass(s):
    return "list"


@dispatch(Callable)  # noqa
def sexpclass(f):
    return "PyCallable"


@dispatch(object)  # noqa
def sexpclass(f):
    return "PyObject"
